@@ -128,15 +128,6 @@ def list_(show_all=False,
             continue
 
         if '_seconds' in schedule[job]:
-            # if _seconds is greater than zero
-            # then include the original back in seconds.
-            # otherwise remove seconds from the listing as the
-            # original item didn't include it.
-            if schedule[job]['_seconds'] > 0:
-                schedule[job]['seconds'] = schedule[job]['_seconds']
-            elif 'seconds' in schedule[job]:
-                del schedule[job]['seconds']
-
             # remove _seconds from the listing
             del schedule[job]['_seconds']
 
@@ -215,7 +206,7 @@ def purge(**kwargs):
                         else:
                             ret['comment'].append('Failed to delete job {0} from schedule.'.format(name))
                             ret['result'] = True
-                        return ret
+
             except KeyError:
                 # Effectively a no-op, since we can't really return without an event system
                 ret['comment'] = 'Event module not available. Schedule add failed.'
