@@ -117,9 +117,25 @@ A REST API for Salt
     static_path : ``/static``
         The URL prefix to use when serving static assets out of the directory
         specified in the ``static`` setting.
-    app
+    enable_sessions : ``True``
+        Enable or disable all endpoints that rely on session cookies. This can
+        be useful to enforce only header-based authentication.
+
+        .. versionadded:: 2017.7.0
+
+    app : ``index.html``
         A filesystem path to an HTML file that will be served as a static file.
         This is useful for bootstrapping a single-page JavaScript app.
+
+        Warning! If you set this option to a custom web application, anything
+        that uses cookie-based authentcation is vulnerable to XSRF attacks.
+        Send the custom ``X-Auth-Token`` header instead and consider disabling
+        the ``enable_sessions`` setting.
+
+        .. versionchanged:: 2017.7.0
+
+            Add a proof-of-concept JavaScript single-page app.
+
     app_path : ``/app``
         The URL prefix to use for serving the HTML file specified in the ``app``
         setting. This should be a simple name containing no slashes.
@@ -465,12 +481,11 @@ import json
 import os
 import signal
 import tarfile
-import time
 from multiprocessing import Process, Pipe
 
 # Import third-party libs
 # pylint: disable=import-error
-import cherrypy
+import cherrypy  # pylint: disable=3rd-party-module-not-gated
 import yaml
 import salt.ext.six as six
 # pylint: enable=import-error
@@ -479,6 +494,7 @@ import salt.ext.six as six
 # Import Salt libs
 import salt
 import salt.auth
+import salt.utils
 import salt.utils.event
 
 # Import salt-api libs
@@ -649,28 +665,6 @@ def salt_auth_tool():
     cherrypy.response.headers['Cache-Control'] = 'private'
 
 
-def cors_handler(*args, **kwargs):
-    '''
-    Check a CORS preflight request and return a valid response
-    '''
-    req_head = cherrypy.request.headers
-    resp_head = cherrypy.response.headers
-
-    ac_method = req_head.get('Access-Control-Request-Method', None)
-
-    allowed_methods = ['GET', 'POST']
-    allowed_headers = ['X-Auth-Token', 'Content-Type']
-
-    if ac_method and ac_method in allowed_methods:
-        resp_head['Access-Control-Allow-Methods'] = ', '.join(allowed_methods)
-        resp_head['Access-Control-Allow-Headers'] = ', '.join(allowed_headers)
-
-        resp_head['Connection'] = 'keep-alive'
-        resp_head['Access-Control-Max-Age'] = '1400'
-
-    return {}
-
-
 def cors_tool():
     '''
     Handle both simple and complex CORS requests
@@ -687,9 +681,33 @@ def cors_tool():
     resp_head['Access-Control-Expose-Headers'] = 'GET, POST'
     resp_head['Access-Control-Allow-Credentials'] = 'true'
 
-    # If this is a non-simple CORS preflight request swap out the handler.
+    # Non-simple CORS preflight request; short-circuit the normal handler.
     if cherrypy.request.method == 'OPTIONS':
-        cherrypy.serving.request.handler = cors_handler
+        ac_method = req_head.get('Access-Control-Request-Method', None)
+
+        allowed_methods = ['GET', 'POST']
+        allowed_headers = [
+            'Content-Type',
+            'X-Auth-Token',
+            'X-Requested-With',
+        ]
+
+        if ac_method and ac_method in allowed_methods:
+            resp_head['Access-Control-Allow-Methods'] = ', '.join(allowed_methods)
+            resp_head['Access-Control-Allow-Headers'] = ', '.join(allowed_headers)
+
+            resp_head['Connection'] = 'keep-alive'
+            resp_head['Access-Control-Max-Age'] = '1400'
+
+        # CORS requests should short-circuit the other tools.
+        cherrypy.response.body = ''
+        cherrypy.response.status = 200
+        cherrypy.serving.request.handler = None
+
+        # Needed to avoid the auth_tool check.
+        if cherrypy.request.config.get('tools.sessions.on', False):
+            cherrypy.session['token'] = True
+        return True
 
 
 # Be conservative in what you send
@@ -751,7 +769,10 @@ def hypermedia_handler(*args, **kwargs):
     cherrypy.response.headers['Content-Type'] = best
     out = cherrypy.response.processors[best]
     try:
-        return out(ret)
+        response = out(ret)
+        if six.PY3:
+            response = salt.utils.to_bytes(response)
+        return response
     except Exception:
         msg = 'Could not serialize the return data from Salt.'
         logger.debug(msg, exc_info=True)
@@ -767,14 +788,17 @@ def hypermedia_out():
     '''
     request = cherrypy.serving.request
     request._hypermedia_inner_handler = request.handler
-    request.handler = hypermedia_handler
+
+    # If handler has been explicitly set to None, don't override.
+    if request.handler is not None:
+        request.handler = hypermedia_handler
 
 
-@functools.wraps
 def process_request_body(fn):
     '''
     A decorator to skip a processor function if process_request_body is False
     '''
+    @functools.wraps(fn)
     def wrapped(*args, **kwargs):  # pylint: disable=C0111
         if cherrypy.request.process_request_body is not False:
             fn(*args, **kwargs)
@@ -799,8 +823,21 @@ def urlencoded_processor(entity):
 
     :param entity: raw POST data
     '''
+    if six.PY3:
+        # https://github.com/cherrypy/cherrypy/pull/1572
+        contents = six.StringIO()
+        entity.fp.read(fp_out=contents)
+        contents.seek(0)
+        body_str = contents.read()
+        body_bytes = salt.utils.to_bytes(body_str)
+        body_bytes = six.BytesIO(body_bytes)
+        body_bytes.seek(0)
+        # Patch fp
+        entity.fp = body_bytes
+        del contents
     # First call out to CherryPy's default processor
     cherrypy._cpreqbody.process_urlencoded(entity)
+    cherrypy._cpreqbody.process_urlencoded(entity)
     cherrypy.serving.request.unserialized_data = entity.params
     cherrypy.serving.request.raw_body = ''
 
@@ -812,7 +849,15 @@ def json_processor(entity):
 
     :param entity: raw POST data
     '''
-    body = entity.fp.read()
+    if six.PY2:
+        body = entity.fp.read()
+    else:
+        # https://github.com/cherrypy/cherrypy/pull/1572
+        contents = six.StringIO()
+        body = entity.fp.read(fp_out=contents)
+        contents.seek(0)
+        body = contents.read()
+        del contents
     try:
         cherrypy.serving.request.unserialized_data = json.loads(body)
     except ValueError:
@@ -828,7 +873,14 @@ def yaml_processor(entity):
 
     :param entity: raw POST data
     '''
-    body = entity.fp.read()
+    if six.PY2:
+        body = entity.fp.read()
+    else:
+        # https://github.com/cherrypy/cherrypy/pull/1572
+        contents = six.StringIO()
+        body = entity.fp.read(fp_out=contents)
+        contents.seek(0)
+        body = contents.read()
     try:
         cherrypy.serving.request.unserialized_data = yaml.safe_load(body)
     except ValueError:
@@ -847,7 +899,14 @@ def text_processor(entity):
 
     :param entity: raw POST data
     '''
-    body = entity.fp.read()
+    if six.PY2:
+        body = entity.fp.read()
+    else:
+        # https://github.com/cherrypy/cherrypy/pull/1572
+        contents = six.StringIO()
+        body = entity.fp.read(fp_out=contents)
+        contents.seek(0)
+        body = contents.read()
     try:
         cherrypy.serving.request.unserialized_data = json.loads(body)
     except ValueError:
@@ -894,6 +953,7 @@ def lowdata_fmt():
 
     This tool requires that the hypermedia_in tool has already been run.
     '''
+
     if cherrypy.request.method.upper() != 'POST':
         return
 
@@ -913,22 +973,28 @@ def lowdata_fmt():
         cherrypy.serving.request.lowstate = data
 
 
-cherrypy.tools.html_override = cherrypy.Tool('on_start_resource',
-        html_override_tool, priority=53)
-cherrypy.tools.salt_token = cherrypy.Tool('on_start_resource',
-        salt_token_tool, priority=55)
-cherrypy.tools.cors_tool = cherrypy.Tool('before_request_body',
-        cors_tool, priority=50)
-cherrypy.tools.salt_auth = cherrypy.Tool('before_request_body',
-        salt_auth_tool, priority=60)
-cherrypy.tools.hypermedia_in = cherrypy.Tool('before_request_body',
-        hypermedia_in)
-cherrypy.tools.lowdata_fmt = cherrypy.Tool('before_handler',
-        lowdata_fmt, priority=40)
-cherrypy.tools.hypermedia_out = cherrypy.Tool('before_handler',
-        hypermedia_out)
-cherrypy.tools.salt_ip_verify = cherrypy.Tool('before_handler',
-        salt_ip_verify_tool)
+tools_config = {
+    'on_start_resource': [
+        ('html_override', html_override_tool),
+        ('salt_token', salt_token_tool),
+    ],
+    'before_request_body': [
+        ('cors_tool', cors_tool),
+        ('salt_auth', salt_auth_tool),
+        ('hypermedia_in', hypermedia_in),
+    ],
+    'before_handler': [
+        ('lowdata_fmt', lowdata_fmt),
+        ('hypermedia_out', hypermedia_out),
+        ('salt_ip_verify', salt_ip_verify_tool),
+    ],
+}
+
+for hook, tool_list in tools_config.items():
+    for idx, tool_config in enumerate(tool_list):
+        tool_name, tool_fn = tool_config
+        setattr(cherrypy.tools, tool_name, cherrypy.Tool(
+            hook, tool_fn, priority=(50 + idx)))
 
 
 ###############################################################################
@@ -942,6 +1008,7 @@ class LowDataAdapter(object):
     exposed = True
 
     _cp_config = {
+        'tools.salt_token.on': True,
         'tools.sessions.on': True,
         'tools.sessions.timeout': 60 * 10,  # 10 hours
 
@@ -955,6 +1022,7 @@ class LowDataAdapter(object):
 
     def __init__(self):
         self.opts = cherrypy.config['saltopts']
+        self.apiopts = cherrypy.config['apiopts']
         self.api = salt.netapi.NetapiClient(self.opts)
 
     def exec_lowstate(self, client=None, token=None):
@@ -1002,6 +1070,7 @@ class LowDataAdapter(object):
             else:
                 yield ret
 
+    @cherrypy.config(**{'tools.sessions.on': False})
     def GET(self):
         '''
         An explanation of the API with links of where to go next
@@ -1109,7 +1178,6 @@ class Minions(LowDataAdapter):
     Convenience URLs for working with minions
     '''
     _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': True,
     })
 
@@ -1229,7 +1297,6 @@ class Minions(LowDataAdapter):
 
 class Jobs(LowDataAdapter):
     _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': True,
     })
 
@@ -1317,13 +1384,13 @@ class Jobs(LowDataAdapter):
                 - 2
                 - 6.9141387939453125e-06
         '''
-        lowstate = [{
-            'client': 'runner',
-            'fun': 'jobs.list_job' if jid else 'jobs.list_jobs',
-            'jid': jid,
-        }]
+        lowstate = {'client': 'runner'}
+        if jid:
+            lowstate.update({'fun': 'jobs.list_job', 'jid': jid})
+        else:
+            lowstate.update({'fun': 'jobs.list_jobs'})
 
-        cherrypy.request.lowstate = lowstate
+        cherrypy.request.lowstate = [lowstate]
         job_ret_info = list(self.exec_lowstate(
             token=cherrypy.session.get('token')))
 
@@ -1332,7 +1399,7 @@ class Jobs(LowDataAdapter):
             ret['info'] = [job_ret_info[0]]
             minion_ret = {}
             returns = job_ret_info[0].get('Result')
-            for minion in returns.keys():
+            for minion in returns:
                 if u'return' in returns[minion]:
                     minion_ret[minion] = returns[minion].get(u'return')
                 else:
@@ -1354,7 +1421,6 @@ class Keys(LowDataAdapter):
     module <salt.wheel.key>` functions.
     '''
 
-    @cherrypy.config(**{'tools.salt_token.on': True})
     def GET(self, mid=None):
         '''
         Show the list of minion keys or detail on a specific key
@@ -1447,6 +1513,10 @@ class Keys(LowDataAdapter):
         Accepts all the same parameters as the :py:func:`key.gen_accept
         <salt.wheel.key.gen_accept>`.
 
+        .. note:: A note about ``curl``
+           Avoid using the ``-i`` flag or HTTP headers will be written and
+           produce an invalid tar file.
+
         Example partial kickstart script to bootstrap a new minion:
 
         .. code-block:: text
@@ -1521,10 +1591,10 @@ class Keys(LowDataAdapter):
         priv_key_file = tarfile.TarInfo('minion.pem')
         priv_key_file.size = len(priv_key)
 
-        fileobj = six.moves.StringIO()
+        fileobj = six.StringIO()
         tarball = tarfile.open(fileobj=fileobj, mode='w')
-        tarball.addfile(pub_key_file, six.moves.StringIO(pub_key))
-        tarball.addfile(priv_key_file, six.moves.StringIO(priv_key))
+        tarball.addfile(pub_key_file, six.StringIO(pub_key))
+        tarball.addfile(priv_key_file, six.StringIO(priv_key))
         tarball.close()
 
         headers = cherrypy.response.headers
@@ -1723,7 +1793,6 @@ class Logout(LowDataAdapter):
     Class to remove or invalidate sessions
     '''
     _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': True,
 
         'tools.lowdata_fmt.on': False,
@@ -1739,9 +1808,87 @@ class Logout(LowDataAdapter):
         return {'return': "Your token has been cleared"}
 
 
+class Token(LowDataAdapter):
+    '''
+    Generate a Salt token from eauth credentials
+
+    Wraps functionality in the :py:mod:`auth Runner <salt.runners.auth>`.
+
+    .. versionadded:: 2017.7.0
+    '''
+    @cherrypy.config(**{'tools.sessions.on': False})
+    def POST(self, **kwargs):
+        r'''
+        .. http:post:: /token
+
+            Generate a Salt eauth token
+
+            :status 200: |200|
+            :status 400: |400|
+            :status 401: |401|
+
+        **Example request:**
+
+        .. code-block:: bash
+
+            curl -sSk https://localhost:8000/token \
+                -H 'Content-type: application/json' \
+                -d '{
+                    "username": "saltdev",
+                    "password": "saltdev",
+                    "eauth": "auto"
+                }'
+
+        **Example response:**
+
+        .. code-block:: http
+
+            HTTP/1.1 200 OK
+            Content-Type: application/json
+
+            [{
+                "start": 1494987445.528182,
+                "token": "e72ca1655d05...",
+                "expire": 1495030645.528183,
+                "name": "saltdev",
+                "eauth": "auto"
+            }]
+        '''
+        for creds in cherrypy.request.lowstate:
+            try:
+                creds.update({
+                    'client': 'runner',
+                    'fun': 'auth.mk_token',
+                    'kwarg': {
+                        'username': creds['username'],
+                        'password': creds['password'],
+                        'eauth': creds['eauth'],
+                    },
+                })
+            except KeyError:
+                raise cherrypy.HTTPError(400,
+                    'Require "username", "password", and "eauth" params')
+
+        return list(self.exec_lowstate())
+
+
 class Run(LowDataAdapter):
     '''
-    Class to run commands without normal session handling
+    Run commands bypassing the :ref:`normal session handling
+    <rest_cherrypy-auth>`
+
+    salt-api does not enforce authorization, Salt's eauth system does that.
+    Local/Runner/WheelClient all accept ``username``/``password``/``eauth``
+    **or** ``token`` kwargs that are then checked by the eauth system. The
+    session mechanism in ``rest_cherrypy`` simply pairs a session with a Salt
+    eauth token and then passes the ``token`` kwarg in automatically.
+
+    If you already have a Salt eauth token, perhaps generated by the
+    :py:func:`mk_token <salt.runners.auth.mk_token>` function in the Auth
+    Runner module, then there is no reason to use sessions.
+
+    This endpoint accepts either a ``username``, ``password``, ``eauth`` trio,
+    **or** a ``token`` kwarg and does not make use of sessions at all.
     '''
     _cp_config = dict(LowDataAdapter._cp_config, **{
         'tools.sessions.on': False,
@@ -1750,16 +1897,13 @@ class Run(LowDataAdapter):
     def POST(self, **kwargs):
         '''
         Run commands bypassing the :ref:`normal session handling
-        <rest_cherrypy-auth>`
+        <rest_cherrypy-auth>` Other than that this URL is identical to the
+        :py:meth:`root URL (/) <LowDataAdapter.POST>`.
 
         .. http:post:: /run
 
-            This entry point is primarily for "one-off" commands. Each request
-            must pass full Salt authentication credentials. Otherwise this URL
-            is identical to the :py:meth:`root URL (/) <LowDataAdapter.POST>`.
-
-            :term:`lowstate` data describing Salt commands must be sent in the
-            request body.
+            An array of :term:`lowstate` data describing Salt commands must be
+            sent in the request body.
 
             :status 200: |200|
             :status 400: |400|
@@ -1782,6 +1926,20 @@ class Run(LowDataAdapter):
                     "eauth": "auto"
                 }]'
 
+        **Or** using a Salt Eauth token:
+
+        .. code-block:: bash
+
+            curl -sS localhost:8000/run \\
+                -H 'Accept: application/x-yaml' \\
+                -H 'Content-type: application/json' \\
+                -d '[{
+                    "client": "local",
+                    "tgt": "*",
+                    "fun": "test.ping",
+                    "token": "<salt eauth token here>"
+                }]'
+
         .. code-block:: http
 
             POST /run HTTP/1.1
@@ -1876,7 +2034,6 @@ class Events(object):
         'tools.encode.encoding': 'utf-8',
 
         # Auth handled manually below
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': False,
 
         'tools.hypermedia_in.on': False,
@@ -1976,11 +2133,31 @@ class Events(object):
             source.onopen = function() { console.info('Listening ...') };
             source.onerror = function(err) { console.error(err) };
             source.onmessage = function(message) {
-              var saltEvent = JSON.parse(message.data);
-              console.info(saltEvent.tag)
-              console.debug(saltEvent.data)
+                var saltEvent = JSON.parse(message.data);
+                console.log(saltEvent.tag, saltEvent.data);
             };
 
+        Note, the SSE stream is fast and completely asynchronous and Salt is
+        very fast. If a job is created using a regular POST request, it is
+        possible that the job return will be available on the SSE stream before
+        the response for the POST request arrives. It is important to take that
+        asynchronity into account when designing an application. Below are some
+        general guidelines.
+
+        * Subscribe to the SSE stream _before_ creating any events.
+        * Process SSE events directly as they arrive and don't wait for any
+          other process to "complete" first (like an ajax request).
+        * Keep a buffer of events if the event stream must be used for
+          synchronous lookups.
+        * Be cautious in writing Salt's event stream directly to the DOM. It is
+          very busy and can quickly overwhelm the memory allocated to a
+          browser tab.
+
+        A full, working proof-of-concept JavaScript appliction is available
+        :blob:`adjacent to this file <salt/netapi/rest_cherrypy/index.html>`.
+        It can be viewed by pointing a browser at the ``/app`` endpoint in a
+        running ``rest_cherrypy`` instance.
+
         Or using CORS:
 
         .. code-block:: javascript
@@ -2073,7 +2250,6 @@ class WebsocketEndpoint(object):
         'tools.encode.encoding': 'utf-8',
 
         # Auth handled manually below
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': False,
 
         'tools.hypermedia_in.on': False,
@@ -2238,7 +2414,6 @@ class WebsocketEndpoint(object):
                         logger.error(
                                 "Error: Salt event has non UTF-8 data:\n{0}"
                                 .format(data))
-                time.sleep(0.1)
 
         parent_pipe, child_pipe = Pipe()
         handler.pipe = parent_pipe
@@ -2301,7 +2476,6 @@ class Webhook(object):
         'tools.lowdata_fmt.on': True,
 
         # Auth can be overridden in __init__().
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': True,
     })
 
@@ -2315,7 +2489,6 @@ class Webhook(object):
                 listen=False)
 
         if cherrypy.config['apiopts'].get('webhook_disable_auth'):
-            self._cp_config['tools.salt_token.on'] = False
             self._cp_config['tools.salt_auth.on'] = False
 
     def POST(self, *args, **kwargs):
@@ -2424,7 +2597,6 @@ class Stats(object):
     exposed = True
 
     _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': True,
         'tools.salt_auth.on': True,
     })
 
@@ -2477,7 +2649,12 @@ class App(object):
             :status 401: |401|
         '''
         apiopts = cherrypy.config['apiopts']
-        return cherrypy.lib.static.serve_file(apiopts['app'])
+
+        default_index = os.path.abspath(os.path.join(
+            os.path.dirname(__file__), 'index.html'))
+
+        return cherrypy.lib.static.serve_file(
+                apiopts.get('app', default_index))
 
 
 class API(object):
@@ -2488,6 +2665,7 @@ class API(object):
         'index': LowDataAdapter,
         'login': Login,
         'logout': Logout,
+        'token': Token,
         'minions': Minions,
         'run': Run,
         'jobs': Jobs,
@@ -2502,7 +2680,15 @@ class API(object):
 
         CherryPy uses class attributes to resolve URLs.
         '''
-        for url, cls in six.iteritems(self.url_map):
+        if self.apiopts.get('enable_sessions', True) is False:
+            url_blacklist = ['login', 'logout', 'minions', 'jobs']
+        else:
+            url_blacklist = []
+
+        urls = ((url, cls) for url, cls in six.iteritems(self.url_map)
+                if url not in url_blacklist)
+
+        for url, cls in urls:
             setattr(self, url, cls())
 
     def _update_url_map(self):
@@ -2520,10 +2706,9 @@ class API(object):
         })
 
         # Enable the single-page JS app URL.
-        if 'app' in self.apiopts:
-            self.url_map.update({
-                self.apiopts.get('app_path', 'app').lstrip('/'): App,
-            })
+        self.url_map.update({
+            self.apiopts.get('app_path', 'app').lstrip('/'): App,
+        })
 
     def __init__(self):
         self.opts = cherrypy.config['saltopts']
