@@ -14,7 +14,7 @@ from collections import defaultdict
 # Import salt libs
 import salt.utils
 import salt.utils.args
-from salt.exceptions import CommandNotFoundError, CommandExecutionError
+from salt.exceptions import CommandNotFoundError, CommandExecutionError, SaltConfigurationError
 from salt.version import SaltStackVersion, __saltstack_version__
 from salt.log import LOG_LEVELS
 
@@ -236,14 +236,28 @@ def memoize(func):
     '''
     Memoize aka cache the return output of a function
     given a specific set of arguments
+
+    .. versionedited:: 2016.3.4
+
+    Added **kwargs support.
     '''
     cache = {}
 
     @wraps(func)
-    def _memoize(*args):
-        if args not in cache:
-            cache[args] = func(*args)
-        return cache[args]
+    def _memoize(*args, **kwargs):
+        str_args = []
+        for arg in args:
+            if not isinstance(arg, six.string_types):
+                str_args.append(str(arg))
+            else:
+                str_args.append(arg)
+        args = str_args
+
+        args_ = ','.join(list(args) + ['{0}={1}'.format(k, kwargs[k]) for k in sorted(kwargs)])
+        if args_ not in cache:
+            cache[args_] = func(*args, **kwargs)
+        return cache[args_]
+
     return _memoize
 
 
@@ -253,6 +267,9 @@ class _DeprecationDecorator(object):
     Takes care of a common functionality, used in its derivatives.
     '''
 
+    OPT_IN = 1
+    OPT_OUT = 2
+
     def __init__(self, globals, version):
         '''
         Constructor.
@@ -266,7 +283,6 @@ class _DeprecationDecorator(object):
         self._exp_version_name = version
         self._exp_version = SaltStackVersion.from_name(self._exp_version_name)
         self._curr_version = __saltstack_version__.info
-        self._options = self._globals['__opts__']
         self._raise_later = None
         self._function = None
         self._orig_f_name = None
@@ -281,11 +297,15 @@ class _DeprecationDecorator(object):
         _args = list()
         _kwargs = dict()
 
-        for arg_item in kwargs.get('__pub_arg', list()):
-            if type(arg_item) == dict:
-                _kwargs.update(arg_item.copy())
-            else:
-                _args.append(arg_item)
+        if '__pub_arg' in kwargs:  # For modules
+            for arg_item in kwargs.get('__pub_arg', list()):
+                if type(arg_item) == dict:
+                    _kwargs.update(arg_item.copy())
+                else:
+                    _args.append(arg_item)
+        else:
+            _kwargs = kwargs.copy()  # For states
+
         return _args, _kwargs
 
     def _call_function(self, kwargs):
@@ -302,7 +322,7 @@ class _DeprecationDecorator(object):
             try:
                 return self._function(*args, **kwargs)
             except TypeError as error:
-                error = str(error).replace(self._function.__name__, self._orig_f_name)  # Hide hidden functions
+                error = str(error).replace(self._function, self._orig_f_name)  # Hide hidden functions
                 log.error('Function "{f_name}" was not properly called: {error}'.format(f_name=self._orig_f_name,
                                                                                         error=error))
                 return self._function.__doc__
@@ -492,33 +512,46 @@ class _WithDeprecated(_DeprecationDecorator):
 
     '''
     MODULE_NAME = '__virtualname__'
-    CFG_KEY = 'use_deprecated'
+    CFG_USE_DEPRECATED = 'use_deprecated'
+    CFG_USE_SUPERSEDED = 'use_superseded'
 
-    def __init__(self, globals, version, with_name=None):
+    def __init__(self, globals, version, with_name=None, policy=_DeprecationDecorator.OPT_OUT):
         '''
         Constructor of the decorator 'with_deprecated'
 
         :param globals:
         :param version:
         :param with_name:
+        :param policy:
         :return:
         '''
         _DeprecationDecorator.__init__(self, globals, version)
         self._with_name = with_name
+        self._policy = policy
 
     def _set_function(self, function):
         '''
         Based on the configuration, set to execute an old or a new function.
         :return:
         '''
-        full_name = "{m_name}.{f_name}".format(m_name=self._globals.get(self.MODULE_NAME, ''),
-                                               f_name=function.__name__)
+        full_name = "{m_name}.{f_name}".format(
+            m_name=self._globals.get(self.MODULE_NAME, '') or self._globals['__name__'].split('.')[-1],
+            f_name=function.__name__)
         if full_name.startswith("."):
             self._raise_later = CommandExecutionError('Module not found for function "{f_name}"'.format(
                 f_name=function.__name__))
 
-        if full_name in self._options.get(self.CFG_KEY, list()):
-            self._function = self._globals.get(self._with_name or "_{0}".format(function.__name__))
+        opts = self._globals.get('__opts__', '{}')
+        use_deprecated = full_name in opts.get(self.CFG_USE_DEPRECATED, list())
+        use_superseded = full_name in opts.get(self.CFG_USE_SUPERSEDED, list())
+        if use_deprecated and use_superseded:
+            raise SaltConfigurationError("Function '{0}' is mentioned both in deprecated "
+                                         "and superseded sections. Please remove any of that.".format(full_name))
+        old_function = self._globals.get(self._with_name or "_{0}".format(function.__name__))
+        if self._policy == self.OPT_IN:
+            self._function = function if use_superseded else old_function
+        else:
+            self._function = old_function if use_deprecated else function
 
     def _is_used_deprecated(self):
         '''
@@ -527,8 +560,14 @@ class _WithDeprecated(_DeprecationDecorator):
 
         :return:
         '''
-        return "{m_name}.{f_name}".format(m_name=self._globals.get(self.MODULE_NAME, ''),
-                                          f_name=self._orig_f_name) in self._options.get(self.CFG_KEY, list())
+        func_path = "{m_name}.{f_name}".format(
+            m_name=self._globals.get(self.MODULE_NAME, '') or self._globals['__name__'].split('.')[-1],
+            f_name=self._orig_f_name)
+
+        return func_path in self._globals.get('__opts__').get(
+            self.CFG_USE_DEPRECATED, list()) or (self._policy == self.OPT_IN
+                                                 and not (func_path in self._globals.get('__opts__', {}).get(
+                                                          self.CFG_USE_SUPERSEDED, list()))), func_path
 
     def __call__(self, function):
         '''
@@ -549,7 +588,8 @@ class _WithDeprecated(_DeprecationDecorator):
             :return:
             '''
             self._set_function(function)
-            if self._is_used_deprecated():
+            is_deprecated, func_path = self._is_used_deprecated()
+            if is_deprecated:
                 if self._curr_version < self._exp_version:
                     msg = list()
                     if self._with_name:
@@ -557,10 +597,11 @@ class _WithDeprecated(_DeprecationDecorator):
                                    'expire in version "{version_name}".'.format(
                                        f_name=self._with_name.startswith("_") and self._orig_f_name or self._with_name,
                                        version_name=self._exp_version_name))
+                        msg.append('Use its successor "{successor}" instead.'.format(successor=self._orig_f_name))
                     else:
-                        msg.append('The function is using its deprecated version and will '
-                                   'expire in version "{version_name}".'.format(version_name=self._exp_version_name))
-                    msg.append('Use its successor "{successor}" instead.'.format(successor=self._orig_f_name))
+                        msg.append('The function "{f_name}" is using its deprecated version and will '
+                                   'expire in version "{version_name}".'.format(f_name=func_path,
+                                                                                version_name=self._exp_version_name))
                     log.warning(' '.join(msg))
                 else:
                     msg_patt = 'The lifetime of the function "{f_name}" expired.'
@@ -582,3 +623,21 @@ class _WithDeprecated(_DeprecationDecorator):
 
 
 with_deprecated = _WithDeprecated
+
+
+def ignores_kwargs(*kwarg_names):
+    '''
+    Decorator to filter out unexpected keyword arguments from the call
+
+    kwarg_names:
+        List of argument names to ignore
+    '''
+    def _ignores_kwargs(fn):
+        def __ignores_kwargs(*args, **kwargs):
+            kwargs_filtered = kwargs.copy()
+            for name in kwarg_names:
+                if name in kwargs_filtered:
+                    del kwargs_filtered[name]
+            return fn(*args, **kwargs_filtered)
+        return __ignores_kwargs
+    return _ignores_kwargs
