@@ -61,18 +61,30 @@ class MasterPillarUtil(object):
 
         # my_runner.py
         tgt = 'web*'
-        pillar_util = salt.utils.master.MasterPillarUtil(tgt, expr_form='glob', opts=__opts__)
+        pillar_util = salt.utils.master.MasterPillarUtil(tgt, tgt_type='glob', opts=__opts__)
         pillar_data = pillar_util.get_minion_pillar()
     '''
     def __init__(self,
                  tgt='',
-                 expr_form='glob',
+                 tgt_type='glob',
                  saltenv=None,
                  use_cached_grains=True,
                  use_cached_pillar=True,
                  grains_fallback=True,
                  pillar_fallback=True,
-                 opts=None):
+                 opts=None,
+                 expr_form=None):
+
+        # remember to remove the expr_form argument from this function when
+        # performing the cleanup on this deprecation.
+        if expr_form is not None:
+            salt.utils.warn_until(
+                'Fluorine',
+                'the target type should be passed using the \'tgt_type\' '
+                'argument instead of \'expr_form\'. Support for using '
+                '\'expr_form\' will be removed in Salt Fluorine.'
+            )
+            tgt_type = expr_form
 
         log.debug('New instance of {0} created.'.format(
             self.__class__.__name__))
@@ -85,7 +97,7 @@ class MasterPillarUtil(object):
             self.opts = opts
         self.serial = salt.payload.Serial(self.opts)
         self.tgt = tgt
-        self.expr_form = expr_form
+        self.tgt_type = tgt_type
         self.saltenv = saltenv
         self.use_cached_grains = use_cached_grains
         self.use_cached_pillar = use_cached_pillar
@@ -93,10 +105,10 @@ class MasterPillarUtil(object):
         self.pillar_fallback = pillar_fallback
         self.cache = salt.cache.factory(opts)
         log.debug(
-            'Init settings: tgt: \'{0}\', expr_form: \'{1}\', saltenv: \'{2}\', '
+            'Init settings: tgt: \'{0}\', tgt_type: \'{1}\', saltenv: \'{2}\', '
             'use_cached_grains: {3}, use_cached_pillar: {4}, '
             'grains_fallback: {5}, pillar_fallback: {6}'.format(
-                tgt, expr_form, saltenv, use_cached_grains, use_cached_pillar,
+                tgt, tgt_type, saltenv, use_cached_grains, use_cached_pillar,
                 grains_fallback, pillar_fallback
             )
         )
@@ -156,7 +168,7 @@ class MasterPillarUtil(object):
                        ','.join(minion_ids),
                         'grains.items',
                         timeout=self.opts['timeout'],
-                        expr_form='list')
+                        tgt_type='list')
         return ret
 
     def _get_live_minion_pillar(self, minion_id=None, minion_grains=None):
@@ -231,14 +243,14 @@ class MasterPillarUtil(object):
         return ret
 
     def _tgt_to_list(self):
-        # Return a list of minion ids that match the target and expr_form
+        # Return a list of minion ids that match the target and tgt_type
         minion_ids = []
         ckminions = salt.utils.minions.CkMinions(self.opts)
-        minion_ids = ckminions.check_minions(self.tgt, self.expr_form)
+        minion_ids = ckminions.check_minions(self.tgt, self.tgt_type)
         if len(minion_ids) == 0:
-            log.debug('No minions matched for tgt="{0}" and expr_form="{1}"'.format(self.tgt, self.expr_form))
+            log.debug('No minions matched for tgt="{0}" and tgt_type="{1}"'.format(self.tgt, self.tgt_type))
             return {}
-        log.debug('Matching minions for tgt="{0}" and expr_form="{1}": {2}'.format(self.tgt, self.expr_form, minion_ids))
+        log.debug('Matching minions for tgt="{0}" and tgt_type="{1}": {2}'.format(self.tgt, self.tgt_type, minion_ids))
         return minion_ids
 
     def get_minion_pillar(self):
@@ -670,7 +682,7 @@ def ping_all_connected_minions(opts):
     else:
         tgt = '*'
         form = 'glob'
-    client.cmd(tgt, 'test.ping', expr_form=form)
+    client.cmd(tgt, 'test.ping', tgt_type=form)
 
 # test code for the ConCache class
 if __name__ == '__main__':
