@@ -465,9 +465,6 @@ class Schedule(object):
             config_dir,
             os.path.dirname(self.opts.get('default_include',
                                           salt.config.DEFAULT_MINION_OPTS['default_include'])))
-        if salt.utils.is_proxy():
-            # each proxy will have a separate _schedule.conf file
-            minion_d_dir = os.path.join(minion_d_dir, self.opts['proxyid'])
 
         if not os.path.isdir(minion_d_dir):
             os.makedirs(minion_d_dir)
@@ -519,7 +516,7 @@ class Schedule(object):
         for job in list(self.opts['schedule'].keys()):
             if job.startswith(name):
                 del self.opts['schedule'][job]
-        for job in self._get_schedule(include_opts=False).keys():
+        for job in self._get_schedule(include_opts=False):
             if job.startswith(name):
                 log.warning('Cannot delete job {0}, '
                             'it`s in the pillar!'.format(job))
@@ -553,7 +550,7 @@ class Schedule(object):
 
         # if enabled is not included in the job,
         # assume job is enabled.
-        for job in data.keys():
+        for job in data:
             if 'enabled' not in data[job]:
                 data[job]['enabled'] = True
 
@@ -888,9 +885,10 @@ class Schedule(object):
             ret['success'] = False
             ret['retcode'] = 254
         finally:
-            # Only attempt to return data to the master
-            # if the scheduled job is running on a minion.
-            if '__role' in self.opts and self.opts['__role'] == 'minion':
+            # Only attempt to return data to the master if the scheduled job is running
+            # on a master itself or a minion.
+            if '__role' in self.opts and self.opts['__role'] in ('master', 'minion'):
+                # The 'return_job' option is enabled by default even if not set
                 if 'return_job' in data and not data['return_job']:
                     pass
                 else:
@@ -918,6 +916,7 @@ class Schedule(object):
                         log.exception("Unhandled exception firing event: {0}".format(exc))
 
             log.debug('schedule.handle_func: Removing {0}'.format(proc_fn))
+
             try:
                 os.unlink(proc_fn)
             except OSError as exc:
@@ -939,7 +938,26 @@ class Schedule(object):
         '''
         Evaluate and execute the schedule
         '''
+
+        def _splay(splaytime):
+            '''
+            Calculate splaytime
+            '''
+            splay_ = None
+            if isinstance(splaytime, dict):
+                if splaytime['end'] >= splaytime['start']:
+                    splay_ = random.randint(splaytime['start'],
+                                            splaytime['end'])
+                else:
+                    log.error('schedule.handle_func: Invalid Splay, '
+                              'end must be larger than start. Ignoring splay.')
+            else:
+                splay_ = random.randint(1, splaytime)
+            return splay_
+
         schedule = self._get_schedule()
+        if not isinstance(schedule, dict):
+            raise ValueError('Schedule must be of type dict.')
         if 'enabled' in schedule and not schedule['enabled']:
             return
         for job, data in six.iteritems(schedule):
@@ -967,10 +985,18 @@ class Schedule(object):
                 )
             if 'name' not in data:
                 data['name'] = job
-            # Add up how many seconds between now and then
-            when = 0
-            seconds = 0
-            cron = 0
+
+            if '_next_fire_time' not in data:
+                data['_next_fire_time'] = None
+
+            if '_splay' not in data:
+                data['_splay'] = None
+
+            if 'run_on_start' in data and \
+                    data['run_on_start'] and \
+                    '_run_on_start' not in data:
+                data['_run_on_start'] = True
+
             now = int(time.time())
 
             if 'until' in data:
@@ -1026,26 +1052,40 @@ class Schedule(object):
                 continue
 
             if True in [True for item in time_elements if item in data]:
-                # Add up how many seconds between now and then
-                seconds += int(data.get('seconds', 0))
-                seconds += int(data.get('minutes', 0)) * 60
-                seconds += int(data.get('hours', 0)) * 3600
-                seconds += int(data.get('days', 0)) * 86400
+                if '_seconds' not in data:
+                    interval = int(data.get('seconds', 0))
+                    interval += int(data.get('minutes', 0)) * 60
+                    interval += int(data.get('hours', 0)) * 3600
+                    interval += int(data.get('days', 0)) * 86400
+
+                    data['_seconds'] = interval
+
+                    if not data['_next_fire_time']:
+                        data['_next_fire_time'] = now + data['_seconds']
+
+                    if interval < self.loop_interval:
+                        self.loop_interval = interval
+
             elif 'once' in data:
-                once_fmt = data.get('once_fmt', '%Y-%m-%dT%H:%M:%S')
-
-                try:
-                    once = datetime.datetime.strptime(data['once'], once_fmt)
-                    once = int(time.mktime(once.timetuple()))
-                except (TypeError, ValueError):
-                    log.error('Date string could not be parsed: %s, %s',
-                            data['once'], once_fmt)
+                if data['_next_fire_time'] and \
+                        data['_next_fire_time'] != now and \
+                        not data['_splay']:
                     continue
 
-                if now != once:
-                    continue
-                else:
-                    seconds = 1
+                if not data['_next_fire_time'] and \
+                        not data['_splay']:
+                    once_fmt = data.get('once_fmt', '%Y-%m-%dT%H:%M:%S')
+                    try:
+                        once = datetime.datetime.strptime(data['once'],
+                                                          once_fmt)
+                        data['_next_fire_time'] = int(
+                            time.mktime(once.timetuple()))
+                    except (TypeError, ValueError):
+                        log.error('Date string could not be parsed: %s, %s',
+                                  data['once'], once_fmt)
+                        continue
+                    if data['_next_fire_time'] != now:
+                        continue
 
             elif 'when' in data:
                 if not _WHEN_SUPPORTED:
@@ -1092,36 +1132,29 @@ class Schedule(object):
                         when = int(time.mktime(when__.timetuple()))
                         if when >= now:
                             _when.append(when)
+
+                    if data['_splay']:
+                        _when.append(data['_splay'])
+
                     _when.sort()
                     if _when:
                         # Grab the first element
                         # which is the next run time
                         when = _when[0]
 
-                        # If we're switching to the next run in a list
-                        # ensure the job can run
-                        if '_when' in data and data['_when'] != when:
-                            data['_when_run'] = True
-                            data['_when'] = when
-                        seconds = when - now
+                        if '_run' not in data:
+                            data['_run'] = True
 
-                        # scheduled time is in the past and the run was not triggered before
-                        if seconds < 0 and not data.get('_when_run', False):
-                            continue
+                        if not data['_next_fire_time']:
+                            data['_next_fire_time'] = when
 
-                        if '_when_run' not in data:
-                            data['_when_run'] = True
+                        if data['_next_fire_time'] < when and \
+                                not data['_run']:
+                            data['_next_fire_time'] = when
+                            data['_run'] = True
 
-                        # Backup the run time
-                        if '_when' not in data:
-                            data['_when'] = when
-
-                        # A new 'when' ensure _when_run is True
-                        if when > data['_when']:
-                            data['_when'] = when
-                            data['_when_run'] = True
-
-                    else:
+                    elif not data.get('_run', False):
+                        data['_next_fire_time'] = None
                         continue
 
                 else:
@@ -1155,91 +1188,82 @@ class Schedule(object):
                             log.error('Invalid date string. Ignoring')
                             continue
                     when = int(time.mktime(when__.timetuple()))
-                    now = int(time.time())
-                    seconds = when - now
 
-                    # scheduled time is in the past and the run was not triggered before
-                    if seconds < 0 and not data.get('_when_run', False):
+                    if when < now and \
+                            not data.get('_run', False) and \
+                            not data['_splay']:
+                        data['_next_fire_time'] = None
                         continue
 
-                    if '_when_run' not in data:
-                        data['_when_run'] = True
+                    if '_run' not in data:
+                        data['_run'] = True
 
-                    # Backup the run time
-                    if '_when' not in data:
-                        data['_when'] = when
+                    if not data['_next_fire_time']:
+                        data['_next_fire_time'] = when
 
-                    # A new 'when' ensure _when_run is True
-                    if when > data['_when']:
-                        data['_when'] = when
-                        data['_when_run'] = True
+                    if data['_next_fire_time'] < when and \
+                            not data['_run']:
+                        data['_next_fire_time'] = when
+                        data['_run'] = True
 
             elif 'cron' in data:
                 if not _CRON_SUPPORTED:
                     log.error('Missing python-croniter. Ignoring job {0}'.format(job))
                     continue
 
-                now = int(time.mktime(datetime.datetime.now().timetuple()))
-                try:
-                    cron = int(croniter.croniter(data['cron'], now).get_next())
-                except (ValueError, KeyError):
-                    log.error('Invalid cron string. Ignoring')
-                    continue
-                seconds = cron - now
+                if data['_next_fire_time'] is None:
+                    # Get next time frame for a "cron" job if it has been never
+                    # executed before or already executed in the past.
+                    try:
+                        data['_next_fire_time'] = int(
+                            croniter.croniter(data['cron'], now).get_next())
+                    except (ValueError, KeyError):
+                        log.error('Invalid cron string. Ignoring')
+                        continue
+
+                    # If next job run is scheduled more than 1 minute ahead and
+                    # configured loop interval is longer than that, we should
+                    # shorten it to get our job executed closer to the beginning
+                    # of desired time.
+                    interval = now - data['_next_fire_time']
+                    if interval >= 60 and interval < self.loop_interval:
+                        self.loop_interval = interval
+
             else:
                 continue
 
-            # Check if the seconds variable is lower than current lowest
-            # loop interval needed. If it is lower than overwrite variable
-            # external loops using can then check this variable for how often
-            # they need to reschedule themselves
-            # Not used with 'when' parameter, causes run away jobs and CPU
-            # spikes.
-            if 'when' not in data:
-                if seconds < self.loop_interval:
-                    self.loop_interval = seconds
             run = False
-
-            if 'splay' in data:
-                if 'when' in data:
-                    log.error('Unable to use "splay" with "when" option at this time. Ignoring.')
+            seconds = data['_next_fire_time'] - now
+            if data['_splay']:
+                seconds = data['_splay'] - now
+            if seconds <= 0:
+                if '_seconds' in data:
+                    run = True
+                elif 'when' in data and data['_run']:
+                    data['_run'] = False
+                    run = True
                 elif 'cron' in data:
-                    log.error('Unable to use "splay" with "cron" option at this time. Ignoring.')
-                else:
-                    if '_seconds' not in data:
-                        log.debug('The _seconds parameter is missing, '
-                                  'most likely the first run or the schedule '
-                                  'has been refreshed refresh.')
-                        if 'seconds' in data:
-                            data['_seconds'] = data['seconds']
-                        else:
-                            data['_seconds'] = 0
-
-            if 'when' in data:
-                # scheduled time is now or in the past, and the run was triggered before
-                if seconds <= 0 and data['_when_run']:
-                    data['_when_run'] = False
+                    # Reset next scheduled time because it is in the past now,
+                    # and we should trigger the job run, then wait for the next one.
+                    data['_next_fire_time'] = None
                     run = True
-            elif 'cron' in data:
-                if seconds == 1:
+                elif seconds == 0:
                     run = True
-            else:
-                if job in self.intervals:
-                    if now - self.intervals[job] >= seconds:
-                        run = True
-                else:
-                    # If run_on_start is True, the job will run when the Salt
-                    # minion start.  If the value is False will run at the next
-                    # scheduled run.  Default is True.
-                    if 'run_on_start' in data:
-                        if data['run_on_start']:
-                            run = True
-                        else:
-                            self.intervals[job] = int(time.time())
-                    else:
-                        run = True
 
-            if run:
+            if '_run_on_start' in data and data['_run_on_start']:
+                run = True
+                data['_run_on_start'] = False
+            elif run:
+                if 'splay' in data and not data['_splay']:
+                    splay = _splay(data['splay'])
+                    if now < data['_next_fire_time'] + splay:
+                        log.debug('schedule.handle_func: Adding splay of '
+                                  '{0} seconds to next run.'.format(splay))
+                        run = False
+                        data['_splay'] = data['_next_fire_time'] + splay
+                        if 'when' in data:
+                            data['_run'] = True
+
                 if 'range' in data:
                     if not _RANGE_SUPPORTED:
                         log.error('Missing python-dateutil. Ignoring job {0}'.format(job))
@@ -1263,7 +1287,7 @@ class Schedule(object):
                                     else:
                                         run = False
                                 else:
-                                    if now >= start and now <= end:
+                                    if start <= now <= end:
                                         run = True
                                     else:
                                         run = False
@@ -1278,30 +1302,13 @@ class Schedule(object):
 
             if not run:
                 continue
-            else:
-                if 'splay' in data:
-                    if 'when' in data:
-                        log.error('Unable to use "splay" with "when" option at this time. Ignoring.')
-                    else:
-                        if isinstance(data['splay'], dict):
-                            if data['splay']['end'] >= data['splay']['start']:
-                                splay = random.randint(data['splay']['start'], data['splay']['end'])
-                            else:
-                                log.error('schedule.handle_func: Invalid Splay, end must be larger than start. \
-                                         Ignoring splay.')
-                                splay = None
-                        else:
-                            splay = random.randint(0, data['splay'])
 
-                        if splay:
-                            log.debug('schedule.handle_func: Adding splay of '
-                                      '{0} seconds to next run.'.format(splay))
-                            if 'seconds' in data:
-                                data['seconds'] = data['_seconds'] + splay
-                            else:
-                                data['seconds'] = 0 + splay
+            miss_msg = ''
+            if seconds < 0:
+                miss_msg = ' (runtime missed ' \
+                           'by {0} seconds)'.format(abs(seconds))
 
-                log.info('Running scheduled job: {0}'.format(job))
+            log.info('Running scheduled job: {0}{1}'.format(job, miss_msg))
 
             if 'jid_include' not in data or data['jid_include']:
                 data['jid_include'] = True
@@ -1343,7 +1350,9 @@ class Schedule(object):
                 if multiprocessing_enabled:
                     proc.join()
             finally:
-                self.intervals[job] = now
+                if '_seconds' in data:
+                    data['_next_fire_time'] = now + data['_seconds']
+                data['_splay'] = None
             if salt.utils.is_windows():
                 # Restore our function references.
                 self.functions = functions
